//	Distributed under the Boost Software License Version 1.0 https://www.boost.org/LICENSE_1_0.txt
//	Copyright Gero Peterhoff

#ifndef BOOST_MATH_QUADMATH_INTEL_MATH_HPP
#define BOOST_MATH_QUADMATH_INTEL_MATH_HPP

#include <boost/config.hpp>
#include <boost/math/quadmath/quadmath_intel.hpp>
//#include <boost/type_traits/endian.hpp>
#if defined(__cpp_lib_bit_cast)
#include <bit>
#endif


/*
#if defined(__STDCPP_FLOAT128_T__)
inline constexpr __float128	nansq(const char*const)	noexcept
{
	return __float128(numeric_limits<std::float128_t>::signaling_NaN());
}
#elif defined(__cpp_lib_bit_cast)
inline constexpr __float128	nansq(const char*const)	noexcept
{
	using array_t = std::array<uint64_t, 2>;

	BOOST_IF_CONSTEXPR (boost::is_little_endian::value)	return bit_cast<__float128>(array_t{0ULL, 9223160930622242816ULL});
	else												return bit_cast<__float128>(array_t{9223160930622242816ULL, 0ULL});
}
#else
inline __float128	nansq(const char*const)	noexcept
{
	using array_t = std::array<uint64_t, 2>;

	BOOST_IF_CONSTEXPR (boost::is_little_endian::value)	return reinterpret_cast<__float128>(array_t{0ULL, 9223160930622242816ULL});
	else												return reinterpret_cast<__float128>(array_t{9223160930622242816ULL, 0ULL});
}
#endif
*/


#if BOOST_INTEL
namespace std
{
//	basic
inline __float128 abs(__float128 x)								noexcept	{ return ::__fabsq(x);				}
inline __float128 fabs(__float128 x)							noexcept	{ return ::__fabsq(x);				}
inline __float128 fmod(__float128 x, __float128 y)				noexcept	{ return ::__fmodq(x, y);			}
inline __float128 remainder(__float128 x, __float128 y)			noexcept	{ return ::__remainderq(x, y);		}
inline __float128 remquo(__float128 x, __float128 y, int* quo)	noexcept	{ return ::__remquoq(x, y, quo);	}
inline __float128 fma(__float128 x, __float128 y, __float128 z)	noexcept	{ return ::__fmaq(x, y, z);			}
inline __float128 fmax(__float128 x, __float128 y)				noexcept	{ return ::__fmaxq(x, y);			}
inline __float128 fmin(__float128 x, __float128 y)				noexcept	{ return ::__fminq(x, y);			}
inline __float128 fdim(__float128 x, __float128 y)				noexcept	{ return ::__fdimq(x, y);			}
inline __float128 nanq(const char* arg)							noexcept	{ return ::__nanq(arg);				}

//	exponential
inline __float128 exp(__float128 x)		noexcept	{ return ::__expq(x);	}
inline __float128 exp2(__float128 x)	noexcept	{ return ::__exp2q(x);	}
inline __float128 expm1(__float128 x)	noexcept	{ return ::__expm1q(x);	}

inline __float128 log(__float128 x)		noexcept	{ return ::__logq(x);	}
inline __float128 log10(__float128 x)	noexcept	{ return ::__log10q(x);	}
inline __float128 log2(__float128 x)	noexcept	{ return ::__log2q(x);	}
inline __float128 log1p(__float128 x)	noexcept	{ return ::__log1pq(x);	}

//	power
inline __float128 pow(__float128 b, __float128 e)	noexcept	{ return ::__powq(b, e);	}
inline __float128 sqrt(__float128 x)				noexcept	{ return ::__sqrtq(x);		}
inline __float128 cbrt(__float128 x)				noexcept	{ return ::__cbrtq(x);		}
inline __float128 hypot(__float128 x, __float128 y)	noexcept	{ return ::__hypotq(x, y);	}
inline __float128 hypot(__float128 x, __float128 y, __float128 z)	noexcept	{ return ::__sqrtq(x*x + y*y + z*z);	}

//	trigonometric
inline __float128 sin(__float128 x)		noexcept	{ return ::__sinq(x);	}
inline __float128 cos(__float128 x)		noexcept	{ return ::__cosq(x);	}
inline __float128 tan(__float128 x)		noexcept	{ return ::__tanq(x);	}

inline __float128 asin(__float128 x)	noexcept	{ return ::__asinq(x);	}
inline __float128 acos(__float128 x)	noexcept	{ return ::__acosq(x);	}
inline __float128 atan(__float128 x)	noexcept	{ return ::__atanq(x);	}

inline __float128 atan2(__float128 y, __float128 x)				noexcept	{ return ::__atan2q(y, x);		}
inline void sincos(__float128 x, __float128* s, __float128* c)	noexcept	{ return ::__sincosq(x, s, c);	}

//	hyperbolic
inline __float128 sinh(__float128 x)	noexcept	{ return ::__sinhq(x);	}
inline __float128 cosh(__float128 x)	noexcept	{ return ::__coshq(x);	}
inline __float128 tanh(__float128 x)	noexcept	{ return ::__tanhq(x);	}

inline __float128 asinh(__float128 x)	noexcept	{ return ::__asinhq(x);	}
inline __float128 acosh(__float128 x)	noexcept	{ return ::__acoshq(x);	}
inline __float128 atanh(__float128 x)	noexcept	{ return ::__atanhq(x);	}

//	error and gamma
inline __float128 erf(__float128 x)		noexcept	{ return ::__erfq(x);		}
inline __float128 erfc(__float128 x)	noexcept	{ return ::__erfcq(x);		}
inline __float128 tgamma(__float128 x)	noexcept	{ return ::__tgammaq(x);	}
inline __float128 lgamma(__float128 x)	noexcept	{ return ::__lgammaq(x);	}

//	round
inline __float128 ceil(__float128 x)		noexcept	{ return ::__ceilq(x);		}
inline __float128 floor(__float128 x)		noexcept	{ return ::__floorq(x);		}
inline __float128 trunc(__float128 x)		noexcept	{ return ::__truncq(x);		}
inline __float128 round(__float128 x)		noexcept	{ return ::__roundq(x);		}
inline long int	  lround(__float128 x)		noexcept	{ return ::__lroundq(x);	}
inline long long int llround(__float128 x)	noexcept	{ return ::__llroundq(x);	}
inline __float128 nearbyint(__float128 x)	noexcept	{ return ::__nearbyintq(x);	}
inline __float128 rint(__float128 x)		noexcept	{ return ::__rintq(x);		}
inline long int	  lrint(__float128 x)		noexcept	{ return ::__lrintq(x);		}
inline long long int llrint(__float128 x)	noexcept	{ return ::__llrintq(x);	}

//	manipulation
inline __float128 frexp(__float128 x, int* exp)			noexcept	{ return ::__frexpq(x, exp);		}
inline __float128 ldexp(__float128 x, int exp)			noexcept	{ return ::__ldexpq(x, exp);		}
inline __float128 modf(__float128 x, __float128* iptr)	noexcept	{ return ::__modfq(x, iptr);		}
inline __float128 scalbn(__float128 x, int exp)			noexcept	{ return ::__scalbnq(x, exp);		}
inline __float128 scalbln(__float128 x, long int exp)	noexcept	{ return ::__scalblnq(x, exp);		}
inline int		  ilogb(__float128 x)					noexcept	{ return ::__ilogbq(x);				}
inline __float128 logb(__float128 x)					noexcept	{ return ::__logbq(x);				}
inline __float128 nextafter(__float128 x, __float128 y)	noexcept	{ return ::__nextafterq(x, y);		}
inline __float128 nexttoward(__float128 x, __float128 y)noexcept	{ return -::__nextafterq(-x, -y);	}
inline __float128 copysign(__float128 x, __float128 y)	noexcept	{ return ::__copysignq(x, y);		}

//	classification and comparison
inline int	fpclassify(__float128 x)	noexcept
{
    	 if (::__isnanq(x))				return FP_NAN;
    else if (::__isinfq(x))				return FP_INFINITE;
    else if (x == 0)					return FP_ZERO;
    else if (::__fabs(x) < FLT128_MIN)	return FP_SUBNORMAL;
    else								return FP_NORMAL;
}

inline bool signbit(__float128 x)		noexcept	{ return ::__signbitq(x);	}
inline bool isfinite(__float128 x)		noexcept	{ return ::__finiteq(x);	}
inline bool isinf(__float128 x)			noexcept	{ return ::__isinfq(x);		}
inline bool isnan(__float128 x)			noexcept	{ return ::__isnanq(x);		}
inline bool	isnormal(__float128 x)		noexcept
{
	return fpclassify(x) == FP_NORMAL;
}
inline bool isunordered(__float128 x, __float128 y)	noexcept
{
	return ::__isnanq(x) || ::__isnanq(y);
}
inline bool isgreater(__float128 x, __float128 y)	noexcept
{
	if (BOOST_UNLIKELY(isunordered(x, y)))	return false;
	else									return x > y;
}
inline bool isgreaterequal(__float128 x, __float128 y)	noexcept
{
	if (BOOST_UNLIKELY(isunordered(x, y)))	return false;
	else									return x >= y;
}
inline bool isless(__float128 x, __float128 y)	noexcept
{
	if (BOOST_UNLIKELY(isunordered(x, y)))	return false;
	else									return x < y;
}
inline bool islessequal(__float128 x, __float128 y)	noexcept
{
	if (BOOST_UNLIKELY(isunordered(x, y)))	return false;
	else									return x <= y;
}
inline bool islessgreater(__float128 x, __float128 y)	noexcept
{
	if (BOOST_UNLIKELY(isunordered(x, y)))	return false;
	else									return (x < y) || (x > y);
}

//	more
inline __float128 j0(__float128 x)			noexcept	{ return ::__j0q(x);	}
inline __float128 j1(__float128 x)			noexcept	{ return ::__j1q(x);	}
inline __float128 jn(int x, __float128 y)	noexcept	{ return ::__jnq(x, y);	}

inline __float128 y0(__float128 x)			noexcept	{ return ::__y0q(x);	}
inline __float128 y1(__float128 x)			noexcept	{ return ::__y1q(x);	}
inline __float128 yn(int x, __float128 y)	noexcept	{ return ::__ynq(x, y);	}



template <> struct numeric_limits<__float128>
{
static constexpr bool				is_specialized		= true;
static constexpr __float128			min()				noexcept	{ return FLT128_MIN;	}
static constexpr __float128			max()				noexcept	{ return FLT128_MAX;	}
static constexpr __float128			lowest()			noexcept	{ return -FLT128_MAX;	}
static constexpr int				digits				= FLT128_MANT_DIG;
static constexpr int				digits10			= FLT128_DIG;
static constexpr int				max_digits10		= 36;
static constexpr bool				is_signed			= true;
static constexpr bool				is_integer			= false;
static constexpr bool				is_exact			= false;
static constexpr int				radix				= 2;
static constexpr __float128			epsilon()			noexcept	{ return FLT128_EPSILON;	}
static constexpr __float128			round_error()		noexcept	{ return 0.5Q;				}
static constexpr int				min_exponent		= FLT128_MIN_EXP;
static constexpr int				min_exponent10		= FLT128_MIN_10_EXP;
static constexpr int				max_exponent		= FLT128_MAX_EXP;
static constexpr int				max_exponent10		= FLT128_MAX_10_EXP;
static constexpr bool				has_infinity		= true;
static constexpr bool				has_quiet_NaN		= true;
static constexpr bool				has_signaling_NaN	= false;//bool(issignalingq(0.0Q));
static constexpr float_denorm_style	has_denorm			= denorm_present;
static constexpr bool				has_denorm_loss		= false;
static constexpr __float128			infinity()			noexcept	{ return HUGE_VALQ;			}
static constexpr __float128			quiet_NaN()			noexcept	{ return std::nanq("");		}
static constexpr __float128			signaling_NaN()		noexcept	{ return std::nansq("");	}
static constexpr __float128			denorm_min()		noexcept	{ return FLT128_DENORM_MIN;	}
static constexpr bool				is_iec559			= true;
static constexpr bool				is_bounded			= true;
static constexpr bool				is_modulo			= false;
static constexpr bool				traps				= false;
static constexpr bool				tinyness_before		= false;
static constexpr float_round_style	round_style			= round_to_nearest;
};
}	//	std

#endif	//	BOOST_INTEL
#endif	//	BOOST_MATH_QUADMATH_INTEL_MATH_HPP
